#include <windows.h>
#include <stdio.h>
#include "3DMath.h"
#include "../include/StdFps.h"

//Function Declarations
void InitGeometry();		//Initialize our actual 3D data
void ReleaseGeometry();		//Release 3D data
void SetupMatrices();		//Setup projection and world matrices
void RotateWorld();			//Rotate the world matrix
void Render();				//Render the current frame
LRESULT WINAPI MsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam); //Windows message handler

//Variables
HWND hWnd;								//Main window handle
HDC TheDC, BackDC;						//Front and back buffers
WORD NumFaces;							//Number of faces
Face *FaceStart;						//Face List
Vertex *VertStart;						//Vertex List
RECT ScreenRect={0,0,640,480};			//Window's client dimensions
Matrix matWorld, matViewProj, matFinal;	//World 3D Matrices
Vector3 CameraNorm;						//Camera look vector
bool WireFrame=true, Polys=true, BackFaceCulling=false; //Toggle options


int WINAPI WinMain(HINSTANCE hInst, HINSTANCE, LPSTR, INT)
{
	//Determine required window size
	RECT WindowRect=ScreenRect;
	AdjustWindowRect(&WindowRect, WS_OVERLAPPEDWINDOW, false);

	//Create the window
	WNDCLASSEX wc = {sizeof(WNDCLASSEX), CS_CLASSDC, MsgProc, NULL, NULL, hInst, LoadIcon(hInst, "MAIN_ICON"), NULL, NULL, NULL, "3D Engine", NULL};
	RegisterClassEx(&wc);
	hWnd = CreateWindow("3D Engine", "3D Engine", WS_OVERLAPPEDWINDOW, 100, 100, WindowRect.right-WindowRect.left, WindowRect.bottom-WindowRect.top, NULL, NULL, hInst, NULL);

	//Create back buffer and get front buffer
	TheDC=GetDC(hWnd);
	BackDC=CreateCompatibleDC(TheDC);
	HBITMAP TheBitmap=CreateCompatibleBitmap(TheDC, ScreenRect.right, ScreenRect.bottom);
	SelectObject(BackDC, TheBitmap);
	DeleteObject(TheBitmap);

	InitGeometry(); //Initialize our actual 3D data

	//Show the window
	ShowWindow(hWnd, SW_SHOWDEFAULT);
	UpdateWindow(hWnd);

	//Main message loop - Keep going until window is closed
	MSG msg;
	ZeroMemory(&msg, sizeof(msg));
	while(msg.message!=WM_QUIT)
	{
	if(PeekMessage(&msg, NULL, 0U, 0U, PM_REMOVE)) //Process all windows messages before rendering more frames
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	else //If there is no windows message, render another frame
		Render();
	}

	ReleaseGeometry(); //Release 3D data

	//Release window data
	ReleaseDC(hWnd, TheDC);
	DeleteObject(TheDC);
	UnregisterClass("D3D Tutorial", wc.hInstance);
	return 0;
}

void InitGeometry() //Initialize our actual 3D data
{
	SetupMatrices(); //Only really need to do the main matricies once

	//Request which file to open and if the user wants to use hardware processing
	printf("File to open [if left blank, Models/plane.b3d will be used]: ");
	char FileName[255];
	gets(FileName);

	//Open the file and read in data
		FILE* FileNum=fopen("Models/plane.b3d", "rb");
		if(!FileNum && !(FileNum=fopen("../Models/plane.b3d", "rb")))
		{
			MessageBox(NULL, "File not found", "File Load Error", MB_OK);
			return;
		}
		//Determine file length
		fseek(FileNum, 0, SEEK_END);
		DWORD FileLength=ftell(FileNum);
		fseek(FileNum, 0, SEEK_SET);
		//Read data
		char *TheData=new char[FileLength];
		fread(TheData, FileLength, 1, FileNum);
		fclose(FileNum);

	//Read data positions
	VertStart=(Vertex*)(TheData+sizeof(WORD));
	FaceStart=(Face*)((char*)VertStart+*(WORD*)TheData*sizeof(Vertex)+sizeof(WORD));
	NumFaces=*(((WORD*)FaceStart)-1);
}

void ReleaseGeometry() //Release 3D data
{
	delete[] ((char*)VertStart-sizeof(WORD));
}

void SetupMatrices() //Setup projection and world matrices
{
	Matrix matView, matProj;
	//Camera does not move
	Vector3 vEyePt(0.0f, 3.0f,-5.0f);
	Vector3 vLookatPt(0.0f, 0.0f, 0.0f);
	Vector3 vUpVec(0.0f, 1.0f, 0.0f);
	MatrixLookAtLH(&matView, &vEyePt, &vLookatPt, &vUpVec);

	//Default projection	
	MatrixPerspectiveFovLH(&matProj, _PI/4, 1.0f, 1.0f, 100.0f);

	//Combined matrix data
	matViewProj=matView*matProj;
	CameraNorm=vEyePt-vLookatPt;
}

void RotateWorld() //Rotate the world matrix
{
	UINT iTime=GetTickCount()%4000;
	Matrix Mat1, Mat2;
	matWorld=(*MatrixRotationX(&Mat1,iTime*(2.0f*_PI)/4000.0f)* *MatrixRotationY(&Mat2,iTime*(2.0f*_PI)/2000.0f));
	matFinal=matWorld*matViewProj;
}

void Render() //Render the current frame
{
	//Clear frame to white and set draw colors
	FillRect(BackDC, &ScreenRect, (HBRUSH)GetStockObject(LTGRAY_BRUSH));
	SelectObject(BackDC, (HBRUSH)GetStockObject(Polys ? WHITE_BRUSH : NULL_BRUSH));
	SelectObject(BackDC, (HBRUSH)GetStockObject(WireFrame? BLACK_PEN: NULL_PEN));

	RotateWorld(); //Rotate the model

	Face *FaceOn=FaceStart, *EndFace=FaceOn+NumFaces; //Go through faces as pointers instead of an array

	Vector4 CameraNormRotate=MatrixVectorMultiply(&matWorld, &CameraNorm); //Multiply world matrix against camera vector so we can more quickly determine which polygons are facing backwards
	for(;FaceOn<EndFace;FaceOn++) //Loop through faces
	{
		//If face is backwards, do not draw (back face culling) - we could store this value per face... but this would be against the point of the time tests
		if(BackFaceCulling && Vec3Dot(&((VertStart[FaceOn->VertexIndices[0]]-VertStart[FaceOn->VertexIndices[2]])*(VertStart[FaceOn->VertexIndices[1]]-VertStart[FaceOn->VertexIndices[2]])), (Vector3*)&CameraNormRotate)<0)
			continue;

		//Calculate projected 2D points
		TwoDCoord TheCoords[3];
		for(int i=0;i<3;i++)
			TheCoords[i]=GetTwoDCoord(VertStart+FaceOn->VertexIndices[i], &matFinal, ScreenRect.right, ScreenRect.bottom);

		//Draw the triangle
		Polygon(BackDC, (POINT*)TheCoords, 3);
	}
	BitBlt(TheDC, 0, 0, ScreenRect.right, ScreenRect.bottom, BackDC, 0, 0, SRCCOPY); //Copy back buffer to front buffer

	//Display the current FPS and toggle options
	static FPS TheFps;
	char WinText[40];
	TheFps.AddFrame();
	sprintf(WinText, "w:%d, p:%d, b:%d, #:%d, fps:%d", WireFrame&1, Polys&1, BackFaceCulling&1, NumFaces, (int)TheFps);

	SetWindowText(hWnd, WinText);
}

LRESULT WINAPI MsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) //Windows message handler
{
	if(msg==WM_DESTROY) //If closing window...
	{
		PostQuitMessage(0);
		return 0;
	}
	else if(msg==WM_CHAR) //Keypress for options
		switch((TCHAR)wParam)
		{
			case 'w':WireFrame=!WireFrame; break; //Toggle wireframe mode
			case 'p':Polys=!Polys;break; //Toggle poly mode
			case 'b':BackFaceCulling=!BackFaceCulling;break; //Toggle poly mode
		}

	//Default message handler
	return DefWindowProc(hWnd, msg, wParam, lParam);
}